"""Handles both nested dictionary and list access

Source: https://stackoverflow.com/questions/14692690/access-nested-dictionary-items-via-a-list-of-keys

Usage --

```
from nested_dict_access_by_key_list import get_by_path, set_by_path, in_nested_path

# retrieve nested dict item
>>> d = {"1": {"2":{"3": "salvador"}}} 
>>> get_by_path(d, ["1", "2", "3"])
'salvador'

# Note that nested retrieval may be dynamic
>>> d = {"1": {"2":{"3": "salvador"}}} 
>>> level_2 = "2"
>>> get_by_path(d, ["1", level_2, "3"])
'salvador'

# Note how nested lists are navigated
>>> d
{'1': {'2': {'3': ['seldon', {'4': 'mallow'}]}}}
>>> get_by_path(d, ["1", "2", "3", 1, "4"])
'mallow'

# set nested dict item (note: can also be done dynamically with variables)
>>> d
{'1': {'2': {'3': 'salvador'}}}
>>> set_by_path(d, ["1", "2"], "mallow")
>>> d
{'1': {'2': 'mallow'}}
```

# Use Python "in" operator on nested dict/list
#
# Note how lists are navigated here. We navigate to dict with
# key '3', then jump to index 1, then back to key "4
>>> d
{'1': {'2': {'3': ['seldon', {'4': 'mallow'}]}}}
>>> in_nested_path(d, ["1", "2", "3", 1, "4"])
True
```

"""

from functools import reduce  # forward compatibility for Python3
import operator


def get_by_path(root, items):
    """Access a nested object in root by item sequence"""
    return reduce(operator.getitem, items, root)


def set_by_path(root, items, value):
    """Set a value in a nested object in root by item sequence"""
    get_by_path(root, items[:-1])[items[-1]] = value


def in_nested_path(root, items):
    """ 'in' equivalent for a nested dict/list structure"""
    try:
        get_by_path(root, items)
    except KeyError:
        return False

    return True



from flask import (Blueprint, make_response, json, redirect, request,
    url_for, abort, current_app)

endpoints_1 = Blueprint("endpoints_1", __name__)

# NOTE: Flask logging is supressed when not in 'debug' mode,
# unless log at the 'error' level

# NOTE: 'current_app' will give us access to the Flask app object


# -- Endpoints --

@endpoints_1.route("/")
def index():
    return redirect(url_for("endpoints_1.flask_skeleton"))


@endpoints_1.route("/flask_skeleton")
def flask_skeleton():
    """Index root/landing endpoint."""
    
    # Can access 'current_app' in order to get app logger.
    current_app.logger.info("Root Endpoint hit.")
    
    # Access database if desired
    data = _get_db().query()

    text_banner = "BLAH"
    resp = make_response(text_banner, 200)
    resp.mimetype = "text/plain"
    return resp

# (Further endpoints)
#       .
#       .
#       .


# -- Internal --


def _get_db():
    """Return handler to database. Can do so because the database
    was attached to the app's config in 
    flask_skeleton.rest_api.__init__.create_app()
    """
    return current_app.config["database"]



import argparse
from importlib import import_module
import multiprocessing as mp
import os
import signal
from yaml import safe_load
import sys
import subprocess
from time import sleep

from flask_skeleton.rest_api import CONFIG_FP_EV, DEFAULT_CONFIG_FP


def _get_argparser():
    """ """
    parser = argparse.ArgumentParser()

    parser.add_argument(
        "-c",
        "--config",
        action="store",
        dest="config_fp",
        default=DEFAULT_CONFIG_FP,
    )

    return parser


def main():
    #mp.set_start_method('spawn')   # May need to do this on MacOS to work on python<=3.7

    parser = _get_argparser()
    args = vars(parser.parse_args())
    args["config_fp"] = os.path.abspath(args["config_fp"])
    CONFIG = safe_load(open(args["config_fp"], 'r'))

    # Declare process handlers so can define sigint hander w/o
    # Python complaining
    flask_app_server = None

    #sigint handler (i.e. kill server if ctrl-c the script in terminal)
    def signal_handler(sig, frame):
        flask_app_server.terminate()
        if flask_app_server.poll() is None:
            sleep(2)
            flask_app_server.kill()
        exit()
    signal.signal(signal.SIGINT, signal_handler)  # Interuption
    signal.signal(signal.SIGTERM, signal_handler) # Termination

    # Flask app needs the config, so set environment variable
    # to where the config is found
    os.environ[CONFIG_FP_EV] = args["config_fp"]

    # Start Flask server
    flask_app_server = subprocess.Popen([
        "gunicorn",
        "-b",
        ":".join([str(CONFIG["rest_api"]["host"]), str(CONFIG["rest_api"]["host"])]),
        "--log-file",
        CONFIG["rest_api"]["log_fp"],
        "--log-level",
        CONFIG["rest_api"]["log_level"],
        "flask_skeleton.rest_api.create_app()"],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE
    )

    while True:
        sleep(1)


if __name__ == "__main__":
    main()



"""

-- Purpose

- I frequently just need to initialize a logger with a handler to a file.
I didnt like having to rewrite all the handler init and config every time.
- Additionally, I made this utility able to also return "fangless" loggers in that
it will return a logger with no filehandler, e.g. no logs will be produced
other than possible terminal output. I found this really useful as many customers
wanted scripts with the ability to toggle logging.

-- Usage

# Create default logger.
#   - Default log level is found at log.DEFAULT_LOG_LEVEL.
#   - Default log file is found at log.DEFAULT_LOG_FP
>>> from log import get_logger
>>> log = get_file_logger()
>>> log.warning("Terminus")    # Will log to default log filepath with default log level mask.

# Create logger with arguments.
>>> log = get_file_logger(fp="path_to_log.log", log_level="DEBUG")
>>> log.info("It pays to be obvious when you are known for subtlety.")

# Create "fangless" logger. (see notes above on purpose)
>>> from log import get_logger
>>> log = get_file_logger(enabled=False)

# To change get_logger() defaults, just edit these global vars in log.py
DEFAULT_LOG_FP = "<program_name>.log"
DEFAULT_LOG_LEVEL = "WARNING"
"""


import logging


DEFAULT_LOG_FP = "<program_name>.log"
DEFAULT_LOG_LEVEL = "WARNING"


def get_file_logger(enabled=True, fp=DEFAULT_LOG_FP, log_level=DEFAULT_LOG_LEVEL):
    """Create a logger that outputs to supplied filepath

    Args:
        enabled (bool): Whether logging is enabled. This function is called
            in scripts that call it with suppied arguments, and allows the
            scripts this utility blindly.
        fp (str): filename of the log
        log_level (str): Python logging level to set
        
    Returns:
        (logging.logger) logging handle
    """
    # no matter what a fangless logger is created (for cleaner code)
    logger = logging.getLogger(fp.split("/")[-1].split(".")[0])
    logger.setLevel(logging.DEBUG)
    formatter = logging.Formatter('%(asctime)s %(levelname)-8s %(message)s \n',
                                  datefmt="%Y-%m-%d %H:%M:%S")
    
    if enabled:
        # set filehandler
        fh = logging.FileHandler(fp, mode='w')
        fh.setFormatter(formatter)
        fh.setLevel(getattr(logging, log_level.upper()))
        logger.addHandler(fh)

    return logger



"""
Function and Class I/O Monitor

A utility for wrapping functions and classes in order to record all input and output to functions/methods.

Use Cases:
- Creating input/output data for unit testing.

Usage:

>>from func_io_monitor import func_io_monitor, class_io_monitor, RECORD_TYPES
>>
>># EXAMPLE: Single Function
>># For this example we want to monitor a single function called 'add'
>>
>>def add(x,y)
>>    return x + y
>>
>># This creates a monitor for 'add()' that will record all input/output to 'add()' in JSON format
>># to a file named "add_monitor.json"
>>add_monitor = func_io_monitor(add, record_type=RECORD_TYPES.json, record_fp="add_monitor.json")
>
>># Now use 'add_monitor' as you would 'add()'
>>add_monitor(1,2)
>>3
>>
>># In 'add_monitor.json', you will see:
>># {"__main__.<none>.add": [{"time": 1597074458.0931141, "in": {"args": [1, 2], "kwargs": {}}, "out": "3"}]}
>>
>>
>># EXAMPLE: Entire Class instance
>>
>> # In this case, we want to monitor the entire class instance, i.e. all its methods
>>class arith():
>>    def __init__(self):
>>        self.t = "test"
>>    def add(self, x, y):
>>        return x + y
>>    def del_(self, x, y):
>>        return x - y 
>>
>>a = arith()
>>
>># Creates a monitor for class methods of instance, that will record all inputs/outputs
>># in python logging format the log file "arith.log"
>>arith_monitor = class_io_monitor(a, record_fp="arith.log", record_type=RECORD_TYPES.log)
>>arith_monitor.add(1,2)
>>3
>>arith_monitor.del_(0, -2)
>>-2
>>
>># In "arith.log", you will see:
>>#2020-08-10 12:05:26 INFO     __main__.arith.add - IN: {'args': (1, 2), 'kwargs': {}} - OUT: 3 
>>#
>>#2020-08-10 12:05:26 INFO     __main__.arith.del_ - IN: {'args': (0, -2), 'kwargs': {}} - OUT: 2
>>
"""


from enum import Enum
import inspect
import json
import logging
import os
import time


class RecordTypes(Enum):
    log = 0
    json = 1

    
RECORD_TYPES = RecordTypes


def func_io_monitor(func, record_type=RECORD_TYPES.log, record_fp=None, recorder=None, input_log_formatter=None, output_log_formatter=None):
    """Creates a monitored I/O version of the function.
    
    Supplied function is not altered.
    
    Args:
        func(func): function or method to wrap
        logger (logger): python logging instance
        record_fp ():
        record_type ():
        input_log_formatter (func): a callable to use to format function input for logging.
        output_log_formatter (func): a callable to use to format function output for logging.

    Returns: (func) monitored function
    """
    if not recorder:
        recorder = _get_recorder(record_type, record_fp)
    
    def io_monitor(*args, **kwargs):
        if "__self__" in dir(func):
            func_class = str(func.__self__).split(" ")[0].strip("<")
            func_key = ".".join([func_class, func.__name__])
        else:
            func_class = "<none>"
            func_key = ".".join([func.__module__,func_class, func.__name__])
        if input_log_formatter:
            input_ = input_log_formatter(*args, **kwargs)
        else:
            # default - json structured str
            input_ = {"args": args, "kwargs": kwargs}
        output = func(*args, **kwargs)
        if output_log_formatter:
            r_output = output_log_formatter(output)
        else:
            # default - convert to str
            r_output = str(output)
        recorder.record(func_key, input_, r_output)
        return output
        
    return io_monitor
    
    
def class_io_monitor(class_instance, record_type=RECORD_TYPES.log, record_fp=None, recorder=None, func_input_log_formatters=None, func_ouput_log_formatters=None):
    """Converts class instance to instance where every component method is I/O monitored.

    The returned class instance has all its methods replaced with monitored versions.

    Args:
        class_instance (obj): class instance to convert

    Returns: (obj) class instance
    """
    # create io monitor methods for every method in class instance
    if not recorder:
        recorder = _get_recorder(record_type, record_fp)
    wrapped_methods = {}
    for attr in dir(class_instance):
        if inspect.ismethod(getattr(class_instance, attr)):
            wrapped_methods[attr] = func_io_monitor(getattr(class_instance, attr), recorder=recorder)
    # replace class instance methods with wrapped monitor versions
    for name, monitor_method in wrapped_methods.items():
        setattr(class_instance, name, monitor_method)  
    return class_instance
      
      
 # -- Internal --


def _get_recorder(record_type, record_fp):
    """ """
    if record_type == RECORD_TYPES.log:
        return _recorder_log(record_fp)
    elif record_type == RECORD_TYPES.json:
        return _recorder_json(record_fp)
    else:
        raise ValueError("Recorder type not found")

        
class _recorder_log:
    """ """
    def __init__(self, fp=None):
        enabled = True if fp else False
        fp = fp if fp else "fangless"
        self.log = _get_logger(fp, enabled=enabled, log_level="INFO")
    def record(self, function_key, input_, output):
        """ """
        self.log.info(f"{function_key} - IN: {input_} - OUT: {output}")
        
 
class _recorder_json:
    """ """
    def __init__(self, fp):
        self.fp = os.path.abspath(fp)
    def record(self, function_key, input_, output):
        """ """
        if os.path.isfile(self.fp):
            e_json = json.load(open(self.fp, 'r'))
        else:
            e_json = {}
        if function_key not in e_json.keys():
            e_json[function_key] = []
        e_json[function_key].append(
            {
                "time": time.time(),
                "in": input_,
                "out": output
            }
        )
        json.dump(e_json, open(self.fp, 'w'))
    

def _get_logger(fp, enabled=True, log_level="INFO"):
    """Create a logger that outputs to supplied filepath
    Args:
        enabled (bool): Whether logging is enabled. This function is called
            in scripts that call it with suppied arguments, and allows the
            scripts this utility blindly.
        fp (str): filename of the log
        log_level (str): Python logging level to set
        
    Returns:
        (logging.logger) logging handle
    """
    # no matter what a fangless logger is created (for cleaner code)
    logger = logging.getLogger(fp.split("/")[-1].split(".")[0])
    logger.setLevel(logging.INFO)
    formatter = logging.Formatter('%(asctime)s %(levelname)-8s %(message)s \n',
                                  datefmt="%Y-%m-%d %H:%M:%S")
    
    if enabled:
        # set filehandler
        fh = logging.FileHandler(fp, mode='w')
        fh.setFormatter(formatter)
        fh.setLevel(getattr(logging, log_level.upper()))
        logger.addHandler(fh)

    return logger
     




from importlib import import_module
import logging, logging.handlers
import os

from flask import Flask
from yaml import safe_load


DEFAULT_CONFIG_FP = "/etc/flask_skeleton/config.yml"
CONFIG_FP_EV = "FLASK_SKELETON_CONFIG"   # environment variable that has config filepath
TEST_FLAG_EV = "TEST_FLAG"    # environment variable that is the test flag
TEST_ENDPOINTS_MODULE_EV = "TEST_BLUEPRINT_MODULE"  # environment variable that holds test Blueprint/endpoints full module name


def create_app(config=None):
    app = Flask("Flask Skeleton")

    # Use gunicorn logging mechanism
    # (REF: https://medium.com/@trstringer/logging-flask-and-gunicorn-the-manageable-way-2e6f0b8beb2f)
    if __name__ != "__main__":
        gunicorn_logger = logging.getLogger("gunicorn.error")
        app.logger.handlers = gunicorn_logger.handlers
        app.logger.setLevel(gunicorn_logger.level)

    # --- Initialize App ---
    CONFIG_FILE = os.environ.get(CONFIG_FP_EV)

    if os.environ.get(TEST_FLAG_EV, default="False") != "True":
        # -- production mode --
        app.logger.critical("Production mode set.")
        app.config["config"] = safe_load(open(CONFIG_FILE, 'r'))

        # Initialize database/backend
        db_module = import_module(app.config["database"]["module"])
        app.config["database"] = getattr(db_module, "DatabaseHandler")(
            logger=app.logger,
            *app.config["module_config"]
        )

        # (any further initialization, maybe components/objects that REST
        # API will use.)
        #
        #

        # Load REST API
        endpoint_modules = []
        for module_name in app.config["endpoints"]:
            module_path = "flask_skeleton.rest_api." + module_name
            endpoint_modules.append(module_path)
            module = import_module(module_path)
            blueprint = getattr(module, module_name)   # Blueprint has same name as module
            app.register_blueprint(blueprint)

        app.logger.critical(f"Endpoints loaded: {endpoint_modules}")

    else:
        # -- test mode --
        app.logger.critical("Test mode set.")

        # Add endpoints from module supplied through env variable
        module = import_module(os.environ[TEST_ENDPOINTS_MODULE_EV])
        module_name = os.environ[TEST_ENDPOINTS_MODULE_EV].split(".")[-1]
        blueprint = getattr(module, module_name)   # Blueprint has same name as module
        app.register_blueprint(blueprint)

    return app



"""
--- Purpose
Compare data structure storage performance of:
  - dict
  - list
  - dataclasses
  - slots
  
--- Usage
First, create a simple object data model in a yaml file.

Supported primary types are:
    - str
    - int
    - float
    - complex
    - bool
    - list
    - tuple
    - set
    - dict

Supported item types (e.g. the subtype used for when the primay type
is one of [list, tuple, set, dict].
    - int
    - float
    - bool

Data Model File (yaml)
```
var1: int
var2: bool
var3: str
var4: float
var5: complex
var6: list:int
var7: tuple:bool
var8: set:float
var9: dict:int
```


>>python dict_vs_list_vs_dataclass_vs_slots.py -d <yaml data file> -n <n instances>
"""

import argparse
from dataclasses import dataclass, make_dataclass
from enum import Enum
from pympler import asizeof
import random
import uuid
from yaml import safe_load


STRING = "It is not the critic who counts; not the man who points out how the strong man stumbles, or where the doer of deeds could have done them better. The credit belongs to the man who is actually in the arena, whose face is marred by dust and sweat and blood; who strives valiantly; who errs, who comes short again and again, because there is no effort without error and shortcoming; but who does actually strive to do the deeds; who knows great enthusiasms, the great devotions; who spends himself in a worthy cause; who at the best knows in the end the triumph of high achievement, and who at the worst, if he fails, at least fails while daring greatly, so that his place shall never be with those cold and timid souls who neither know victory nor defeat."
STRING_SIZE = 10
INT_RANGE = (0,10000000)
FLOAT_RANGE = (0, 1000000)
TUPLE_SIZE = 10
LIST_SIZE = 10
DICT_SIZE = 10
SET_SIZE = 10


class TYPES(Enum):
    STR = str
    BOOL = bool
    INT = int
    FLOAT = float
    COMPLEX = complex
    LIST = list
    TUPLE = tuple
    DICT = dict
    SET = set


class ITEM_TYPES(Enum):
    BOOL = bool
    INT = int
    FLOAT = float


def _random_item_values(item_type, size):
    """
    NOTE: no checking of item_type, assumes is one of ITEM_TYPES
    """
    if item_type == ITEM_TYPES.INT:
        l = [random.randint(INT_RANGE[0], INT_RANGE[1]) for i in range(size)]
    elif item_type == ITEM_TYPES.BOOL:
        l = [random.choice([True, False]) for i in range(size)]
    elif item_type == ITEM_TYPES.FLOAT:
        l = [random.uniform(FLOAT_RANGE[0], FLOAT_RANGE[1]) for i in range(size)]
    else:
        raise ValueError(f"Item type {item_type} not supported.")
    return l


def _check_and_get_types(type_str):
    """ """
    if ":" in type_str:
        primary_type, item_type  = type_str.split(":")
    else:
        primary_type = type_str
        item_type = None
    try:
        primary_type = getattr(TYPES, primary_type.upper())
    except ValueError:
        raise Warning(f"Data field type {primary_type} is not supported as a primary field type.")
        primary_type = None
    if item_type:
        try:
            item_type = getattr(ITEM_TYPES, item_type.upper())
        except ValueError:
            raise Warning(f"Data field type {item_type} is not supported as a item field type.")
    return primary_type, item_type


def _random_data_fields(data_model):
    """

    Args:
        data_model(dict): data field names and types

    Returns: (dict)
    """
    fields_values = {}
    for name, type_ in data_model.items():
        primary_type, item_type = _check_and_get_types(type_)
        if not primary_type:
            continue

        if primary_type == TYPES.STR:
            v = STRING[0:STRING_SIZE]
        elif primary_type == TYPES.INT:
            v = random.randint(INT_RANGE[0], INT_RANGE[1])
        elif primary_type == TYPES.BOOL:
            v = random.choice([True, False])
        elif primary_type == TYPES.FLOAT:
            v = random.uniform(FLOAT_RANGE[0], FLOAT_RANGE[1])
        elif primary_type == TYPES.COMPLEX:
            v = TYPES.COMPLEX.value(
                random.randint(INT_RANGE[0], INT_RANGE[1]), 
                random.randint(INT_RANGE[0], INT_RANGE[1]))
        elif primary_type == TYPES.LIST:
            v = TYPES.LIST.value(_random_item_values(item_type, LIST_SIZE))
        elif primary_type == TYPES.TUPLE:
            v = TYPES.TUPLE.value(_random_item_values(item_type, TUPLE_SIZE))
        elif primary_type == TYPES.SET:
            v = TYPES.SET.value(_random_item_values(item_type, SET_SIZE))
        elif primary_type == TYPES.DICT:
            v = {i: item_type for i in range(DICT_SIZE)}

        fields_values[name] = v

    return fields_values


def _get_argparser():
    """to organize and clean format argparser args"""
    parser = argparse.ArgumentParser()

    parser.add_argument(
        "-d",
        "--data-model",
        action="store",
        dest="data_model",
        required=True,
        help="yaml file of data model to use"
    )

    parser.add_argument(
        "-n",
        action="store",
        dest="n",
        default=10000,
        help="count of data structures to create"
    )
    return parser
    
    
def main():
    parser = _get_argparser()

    # parse all args and put in dict
    args = vars(parser.parse_args())
    args["n"] = int(args["n"])

    data_model = safe_load(open(args["data_model"], 'r'))
    DataClass = make_dataclass("MK", data_model)
    SlotsClass =  type("MKS", (object, ), {"__slots__": list(data_model.keys())})

    dicts = []
    lists = []
    slots = []
    data_classes = []
    for _ in range(args['n']):
        fields_values = _random_data_fields(data_model)
        dicts.append(fields_values)
        lists.append(list(fields_values.values()))
        data_classes.append(DataClass(**fields_values))
        sc = SlotsClass()
        for k,v in fields_values.items():
            setattr(sc, k, v)
        slots.append(sc)

    print(f"n is {args['n']}")
    print(f"Size of data classes: {asizeof.asizeof(data_classes)} bytes")
    print(f"Size of slots classes: {asizeof.asizeof(slots)} bytes")
    print(f"Size of lists: {asizeof.asizeof(lists)} bytes")
    print(f"Size of dicts: {asizeof.asizeof(dicts)} bytes")


if __name__ == "__main__":
    main()



import os
import pytest
import yaml

import flask_skeleton


@pytest.fixture(scope="module")
def tests_dir():
    """Test directory"""
    return os.path.join(flask_skeleton.__path__[0], "tests")

@pytest.fixture(scope="module")
def test_config(tests_dir):
    """Load test configuration."""
    fp = os.path.join(tests_dir, "test_config.yml")
    return yaml.safe_load(open(fp, 'r'))



import pathlib
from setuptools import setup, find_packages


PARENT_DIR = pathlib.Path(__file__).parent
README = (PARENT_DIR / "README.md").read_text()

setup(
    name="junkdrawer",
    version="0.0.1",
    license="MIT",
    description="Useful Python utilities and patterns for a career Python developer.",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/elegantmoose/junkdrawer",
    packages=find_packages(),
    install_requires=[
        "flask",
        "gunicorn",
        "pympler",
        "pytest",
        "pyyaml",
    ]
)



"""A clean and organized way to have an argument parser for a script"""


import argparse


def _get_argparser():
    """to organize and clean format argparser args"""
    parser = argparse.ArgumentParser()

    parser.add_argument(
        "--arg_1",
        action="store",
        dest="arg_1",
        default="default_arg_1",
        help="help message"
    )

    parser.add_argument(
        "--arg_2",
        action="store",
        dest="arg_2",
        default="default_arg_2",
        help="help message"
    )

    #         .
    # (and any more arguments)
    #         .

    # mutuall exclusive args

    mutually_exclusive_args = parser.add_mutually_exclusive_group()

    mutually_exclusive_args.add_argument(
        "--me_arg_1",
        action="store",
        dest="me_arg_1",
        default="default_me_arg_1",
        help="help message"    
    )

    mutually_exclusive_args.add_argument(
        "--me_arg_2",
        action="store",
        dest="me_arg_2",
        default="default_me_arg_2",
        help="help message"    
    )

    #         .
    # (and any more mutually exlusive arguments)
    #         .

    # catch all 'vars' argument for user to list any other field/values parameters
    parser.add_argument(
        "--vars",
        action="store",
        dest="vars",
        nargs="*",
        default=None,
        help="Any additional variables that do not have explicit command line options. Supply as a list in form: '--vars var1=val1 var2=val2 ..."
    )

    return parser


def main():

    parser = _get_argparser()

    # parse all args and put in dict
    options = vars(parser.parse_args())

    #         .
    # Rest of script
    #         .


    return


if __name__ == "__main__":
        main()




"""A filter function that can work on any JSON-like data structure
(i.e. lists, dicts, and values). The filter function takes a list of the JSON-like
data instances and then applies filters (specified by simple filter format) to them.
Filters can be supplied to filter out matching instances or filter in matching instances.
The filters can be for string matches, substring check (via python "in"), re.search pattern
or re.match pattern. Filters may also be flagged to be applied separately as a set
(i.e. compound filter).

"""

import copy
import operator
from functools import reduce
import re


def get_by_path(root, items):
    """Access a nested object in root by item sequence.

    Source: https://stackoverflow.com/questions/19589233/reduce-function-with-three-parameters
    """
    return reduce(operator.getitem, items, root)


def in_nested_path(root, items):
    """ 'in' equivalent for a nested dict/list structure"""
    try:
        get_by_path(root, items)
    except KeyError:
        return False

    return True


def set_by_path(root, items, value):
    """Set a vlaue in a nested object in root by item sequence"""
    get_by_path(root, items[:-1])[items[-1]] = value


def filter_data(data, filters, mode="inclusive", compound=True):
    """Filter in or out data instances based on supplied filters

    Can handle fields in nested dict/list(s), at any level

    Deepcopies data so that the data returned is entirely independent set

    Filter format:

    filter_1 = {
        "type": "<type>",
        "field": "<field>",
        "value": "<value>",
        "replace": "<replacement_value>"
    }

    where:

    <type> may be one of "match", "in", "re-match", "re-search", "replace". <type> specifies
    the type of filter mechanism to be applied. "match" is exact string match, "in" is the
    application of the pythonic "in" for substings, and "re-match" and "re-search" apply the
    corresponding python regex function.

    <field> is the field to apply the filter to. The field can also be a nested
    field. For example, "field1.field2.field3" will apply the filter to the data_structure[field1][field2][field3]
    and "field1.[2].field2" would apply the filter to data_structure[field1][2][field2], where the path
    component "[2]" denotes an index in a list.

    <value> is the value of the filter to be evaluated against the field value.
    For filters of type "re-match" and "re-search", these values can be anything
    within the python regex pattern string spec

    <replace> is the value to replace the actual <value> that was matched, if the
    filter type is "replace"

    Args:
        data (list): list of JSON-like data structures. Each item in the list
          should be a JSON-like (i.e. dict and list based) data structure
        filters (list): list of filters, where each fitler is a dict of the
          above specified form
        mode (str): the mode in which to apply the filters to the data. "include"
          specifies that all data instances that match the filter(s) will be
          included in the returned data. "exclude" specifies that all data instances
          that match the filter(s) will be excluded from the returned data. If one of
          the filters is of type 'replace', then this flag is ignored.
        compound (bool): flag to specify whether the filters should be applied
          together (as one compound filter) or separately. If one of the filters
          is of type 'replace', then this flag is ignored.

    Returns: (list) data instances that matched filter(s)
    """
    results = []
    data = copy.deepcopy(data)  # modify locally
    filters = copy.deepcopy(filters)  # modify locally

    if not isinstance(data, list):
        data = [data]  # make list so can iterate regardless

    if not isinstance(filters, list):
        filters = [filters]

    # expand nested field paths to a list (to work with access functions)
    for filt in filters:
        filt["field"] = filt["field"].split(".")

    # Any field path that has the "[#]" for are interpreted as list
    # indexes (instead of keys), so convert them to integers so nested
    # access works correctly
    for filt in filters:
        for idx, path_component in enumerate(filt["field"]):
            if re.match("\[[0-9]+\]", path_component):
                filt["field"][idx] = int(path_component.strip("[]"))

    for instance in data:
        match_count = 0
        replace_flag = False
        match = False

        for filt in filters:
            if not in_nested_path(instance, filt["field"]):
                continue

            if filt["type"] == "match":
                if get_by_path(instance, filt["field"]) == filt["value"]:
                    match_count += 1
                    if not compound:
                        break
            elif filt["type"] == "in":
                if filt["value"] in get_by_path(instance, filt["field"]):
                    match_count += 1
                    if not compound:
                        break
            elif filt["type"] == "re-match":
                if re.match(filt["value"], get_by_path(instance, filt["field"])):
                    match_count += 1
                    if not compound:
                        break
            elif filt["type"] == "re-search":
                if re.search(filt["value"], get_by_path(instance, filt["field"])):
                    match_count += 1
                    if not compound:
                        break
            elif filt["type"] == "replace":
                replace_flag = True
                instance_value = get_by_path(instance, filt["field"])
                matches = re.findall(filt["value"], instance_value)
                if matches:
                    for match in matches:
                        if type(match) is tuple:
                            # multi group regular expression, just going to use first one
                            match = match[0]
                        instance_value = instance_value.replace(match, filt["replace"])
                    set_by_path(instance, filt["field"], instance_value)
                    match_count += 1

        if replace_flag:
            # if one of the filters was of type==replace, 'compound' and
            # 'exclusive/inclusive' options are ignored as can no longer apply.
            # All data instances are returned no matter what (albeit some possibly modified)
            results.append(instance)
        else:
            # Determine if match occured based on filters matched and whether
            # the filters are specified as a compound filter
            if compound:
                if match_count == len(filters):
                    match = True
            elif not compound:
                if match_count > 0:
                    match = True

            if match and mode == "inclusive":
                # inclusive filter mode and a match, so add
                results.append(instance)
            elif not match and mode == "exclusive":
                # exclusive filter mode and not a match, so add
                results.append(instance)

    return results


